//! Command-line parser for `sq cert lint`.

use clap::Args;

use crate::cli::examples::*;
use crate::cli::types::ClapData;
use crate::cli::types::FileOrStdout;
use crate::cli::types::cert_designator::*;

/// Checks for and optionally repairs OpenPGP certificates that use
/// SHA-1.
#[derive(Debug, Args)]
#[clap(
    about = "Check certificates for issues",
    long_about = "Check certificates for issues

`sq cert lint` checks the supplied certificates for the following \
SHA-1-related issues:

  - Whether a certificate revocation uses SHA-1.

  - Whether the current self signature for a non-revoked User ID uses \
    SHA-1.

  - Whether the current subkey binding signature for a non-revoked, \
    live subkey uses SHA-1.

  - Whether a primary key binding signature (\"backsig\") for a \
    non-revoked, live subkey uses SHA-1.

Diagnostics are printed to stderr.  At the end, some statistics are \
shown.  This is useful when examining a keyring.  If `--fix` is \
specified and at least one issue could be fixed, the fixed \
certificates are printed to stdout.

This tool does not currently support smart cards.  But, if only the \
subkeys are on a smart card, this tool may still be able to partially \
repair the certificate.  In particular, it will be able to fix any \
issues with User ID self signatures and subkey binding signatures for \
encryption-capable subkeys, but it will not be able to generate new \
primary key binding signatures for any signing-capable subkeys.
",
    after_help = EXAMPLES,
)]
pub struct Command {
    /// Attempts to fix certificates, when possible.
    #[arg(long)]
    pub fix: bool,

    /// When fixing a certificate, the fixed certificate is exported
    /// without any secret key material.  Using this switch causes any
    /// secret key material to also be exported.
    #[arg(long)]
    pub export_secret_keys: bool,

    /// If set, outputs a list of fingerprints, one per line, of
    /// certificates that have issues.  This output is intended for
    /// use by scripts.
    ///
    /// This option implies `--quiet`. If you also specify `--fix`,
    /// errors will still be printed to stderr, and fixed certificates
    /// will still be emitted to stdout.
    #[arg(long)]
    pub list_keys: bool,

    #[command(flatten)]
    pub certs: CertDesignators<FileCertUserIDEmailDomainGrepArgs,
                               NoPrefix,
                               NoOptions>,

    #[clap(
        long,
        value_name = FileOrStdout::VALUE_NAME,
        help = "Write to the specified FILE.  If not specified, and the \
                certificate was read from the certificate store, imports the \
                modified certificate into the cert store.  If not specified, \
                and the certificate was read from a file, writes the modified \
                certificate to stdout.",
    )]
    pub output: Option<FileOrStdout>,
    #[clap(
        long = "binary",
        help = "Emit binary data",
    )]
    pub binary: bool,
}

const EXAMPLES: Actions = Actions {
    actions: &[
        Action::Setup(Setup {
            command: &[
                "sq", "toolbox", "keyring", "merge",
                "--output=certs.pgp",
                "bob.pgp", "romeo.pgp",
            ],
        }),

        Action::Setup(Setup {
            command: &[
                "sq", "key", "import", "alice-secret.pgp",
            ],
        }),

        Action::Example(Example {
            comment: "\
Gather statistics on the certificates in a keyring.",
            command: &[
                "sq", "cert", "lint",
                "--file", "certs.pgp",
            ],
        }),

        Action::Example(Example {
            comment: "\
Get a list of certificates with issues.",
            command: &[
                "sq", "cert", "lint",
                "--list-keys",
                "--file", "certs.pgp",
            ],
        }),

        Action::Example(Example {
            comment: "\
Fix a key with known problems.",
            command: &[
                "sq", "key", "export",
                "--cert", "EB28F26E2739A4870ECC47726F0073F60FD0CBF0",
                "|", "sq", "cert", "lint", "--fix", "--file=-",
                "|", "sq", "cert", "import"
            ],
        }),
    ],
};
test_examples!(sq_cert_lint, EXAMPLES);
